/*Copyright (C) 2005 Matteo Lucarelli - matteo@matteolucarelli.net
* 
*This program is free software; you can redistribute it and/or
*modify it under the terms of the GNU General Public License
*as published by the Free Software Foundation; either version 2
*of the License, or (at your option) any later version.
*
*This program is distributed in the hope that it will be useful,
*but WITHOUT ANY WARRANTY; without even the implied warranty of
*MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*GNU General Public License for more details.
*
*You should have received a copy of the GNU General Public License
*along with this program; if not, write to the Free Software
*Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
******************************************************************************/

#define PROGVERS "1.6"
#define AUTHORINFO ":(c)2005 Matteo Lucarelli [DiffView: (C)2005 Michael Sweet]"
#define PROGNAME "xrecursedif"

#include <getopt.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <FL/Fl.H>
#include <FL/fl_ask.H>
#include <FL/fl_draw.H>

#ifdef WIN32
#include <windows.h>
#else
#include <signal.h>
#endif

#include "xrdstrings.h"
#include "UIdiff.h"
#include "UImain.h"
#include "FileInfo.h"

#define MAXSTRING 2048
#define FILTERMAX 1024

UImain* UI=NULL;
FileInfo** fiAry=NULL;
int fiAryCount=0;
UIdiff* wDiff=NULL;

bool onlydiff;
bool nocase;
bool projmode;

char* filter=NULL;
char* filterAry[FILTERMAX];   // TODO dynamic
char* filterOAry[FILTERMAX];   // TODO dynamic
int filterCount=0;
int filterOCount=0;

// arrays to store view positions
int FileLine[MAXSTRING];  // TODO dynamic
int FoundLine[MAXSTRING]; // TODO dynamic

void print_usage (int exit_code)
{
	printf (" xRecurseDiff is a recursive diff frontend\n"
		" Searches in project or library directory for file\n"
		" having the same name and compare them using diff\n");
	printf (" Usage: %s [OPTIONS] [PATH1] [PATH2]\n",PROGNAME);
	printf ("  -h  --help     Display this usage information.\n"
		"  -V  --version  Display version informations.\n"
		" Optional PATHs are written in file selection field.\n"
		" Released under the terms of GNU General Public Licence\n"
		" Download last version from www.matteolucarelli.net\n");

	exit (exit_code);
}

// scan directory
void scandir(const char* dir,bool recurse,char* subdir,bool project)
{
	int i;
	bool ret;
	bool passfilter;
	char newsub[MAXSTRING]; // TODO dynamic
	char newdir[MAXSTRING]; // TODO dynamic
	char diffc[MAXSTRING];  // TODO dynamic
	char fname[MAXSTRING];  // TODO dynamic
	
	// check interface
	if (!Fl::check()) exit(0);
	
	// try to open project directory
#ifdef WIN32
	HANDLE hList;
	TCHAR szDir[MAXSTRING]; // TODO dynamic
	WIN32_FIND_DATA FileData;
    	sprintf(szDir, "%s/*", dir);
	hList = FindFirstFile(szDir, &FileData);
	    if (hList == INVALID_HANDLE_VALUE){ 
#else
	DIR* pdir=opendir(dir);
	if (!pdir){
#endif
		fl_alert("%s %s",XRDS_ERROROPENINGDIR,dir);
		return;
	}
	
	// scan directory
#ifdef WIN32
	do{
		strncpy(fname,FileData.cFileName,MAXSTRING);
#else
	struct dirent *ep;
	while (ep = readdir(pdir)){
		strncpy(fname,ep->d_name,MAXSTRING);
#endif
		if (!strcmp(fname,".")) continue;
		if (!strcmp(fname,"..")) continue;
		
		// apply exclude filter
		passfilter=true;
		for (i=0;i<filterOCount;i++)
			if (strstr(fname,filterAry[i])||strstr(subdir,filterAry[i])) passfilter=false;
		if (!passfilter) continue;
		
		// for dirs
#ifdef WIN32
		if (((FileData.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)!=0)&&recurse){
#else
		if ((ep->d_type==DT_DIR)&&recurse){
#endif
			snprintf(newdir,MAXSTRING,"%s/%s",dir,fname);
			snprintf(newsub,MAXSTRING,"%s%s/",subdir,fname);
			scandir(newdir,recurse,newsub,project);
		}
		// for regular files
#ifdef WIN32
		else{
#else
		if (ep->d_type==DT_REG){
#endif
			if (project){ // we're scanning project dir
				// apply filter
				passfilter=(filterCount==0);
				for (i=0;i<filterCount;i++)
					if (strstr(fname,filterAry[i])) passfilter=true;
				
				// store file infos
				if (passfilter){
					fiAry=(FileInfo**)realloc(fiAry,sizeof(FileInfo*)*(fiAryCount+1));
					fiAry[fiAryCount]=new FileInfo(fname,dir,subdir);
					fiAryCount++;
				}
			}else{ // we're scanning library dir
				// scan fiAry name for omonims
				for (i=0;i<fiAryCount;i++){
					// name are equal?
					if (nocase) ret=(strcasecmp(fname,fiAry[i]->m_name)==0);
					else ret=(strcmp(fname,fiAry[i]->m_name)==0);
					
					if (projmode){
						// relative path are equal?
						if (nocase) ret=ret&&(strcasecmp(subdir,fiAry[i]->m_subdir)==0);
						else ret=ret&&(strcmp(subdir,fiAry[i]->m_subdir)==0);						
					}
					if (ret){
#ifdef WIN32
						// diff command
						snprintf(diffc,MAXSTRING,"diff.exe -q \"%s/%s\" \"%s/%s\"",
							dir,fname,fiAry[i]->m_basedir,fiAry[i]->m_name);

						// windows command execution is a little bit(!) verbose
						STARTUPINFO    si;
						PROCESS_INFORMATION  pi;
						DWORD  tstat;
						GetStartupInfo(&si);
						si.wShowWindow=SW_HIDE;
						if (CreateProcess("diff.exe", diffc,NULL,NULL,FALSE,0,NULL,NULL,&si,&pi)==0){
							fl_alert("%s",XRDS_ERRORNODIFF);
							exit(1); 
						}
						WaitForSingleObject(pi.hProcess, INFINITE);
						GetExitCodeProcess(pi.hProcess, &tstat);						
						ret=(tstat==0);
#else
						// diff command
						snprintf(diffc,MAXSTRING,"diff -q \"%s/%s\" \"%s/%s\" &>/dev/null",
							dir,fname,fiAry[i]->m_basedir,fiAry[i]->m_name);
						ret=((system(diffc)==0));
						// TODO control if diff exixsts
#endif
						// store result
						fiAry[i]->AddFound(fname,dir,subdir,ret);

					}
				}
			}
		}
#ifdef WIN32
	} while (FindNextFile(hList, &FileData)!=0);
	FindClose(hList);
#else
	}
	closedir (pdir);
#endif
	return;
}

// parse filter string to array
void parsefilter(const char *fstring)
{
	filterCount=0;
	filterOCount=0;
	
	free(filter);
	filter=(char*)malloc(strlen(fstring)+1);
	strcpy(filter,fstring);
	
	char* fp=filter;
	
	while (*fp!=0){
		while ((*fp==' ')||(*fp==',')||(*fp==';')) *fp++=0;
		if (*fp!=0){ 
			if (*fp=='-') filterAry[filterOCount++]=(fp+1);
			else filterAry[filterCount++]=fp;
			
			if (filterCount==FILTERMAX) filterCount=FILTERMAX-1;
			if (filterOCount==FILTERMAX) filterOCount=FILTERMAX-1;
		}
		while ((*fp!=0)&&(*fp!=' ')&&(*fp!=',')&&(*fp!=';')) fp++;
	}
}

// clear window and array
void clearfilelist()
{
	(UI->filelist)->clear();
	for (int i=0; i<fiAryCount; i++) delete fiAry[i];
	free(fiAry);
	fiAry=NULL;
	fiAryCount=0;
}

// output search results in fl_browser filelist 
void printresult()
{
	int i,j;
	char line[MAXSTRING]; // TODO dynamic
	bool passfilter;
	
	char dir1[MAXSTRING];
	char dir2[MAXSTRING];
	
	if (projmode){
		snprintf(dir1,MAXSTRING,"%s-1/",XRDS_PROJECTDIR);
		snprintf(dir2,MAXSTRING,"%s-2/",XRDS_PROJECTDIR);
	}else{
		snprintf(dir1,MAXSTRING,"%s/",XRDS_PROJECTDIR);
		snprintf(dir2,MAXSTRING,"%s/",XRDS_LIBDIR);
	}	
	
	for (i=0; i<fiAryCount; i++){
	
		if (fiAry[i]->m_FoundCount==0) {
			snprintf(line,MAXSTRING,"@B50%s%s%s %s",
				dir1,
				fiAry[i]->m_subdir,
				fiAry[i]->m_name,
				XRDS_NOTFOUND);
			if (!onlydiff){
				(UI->filelist)->add(line);
				FileLine[(UI->filelist)->size()]=i;
				FoundLine[(UI->filelist)->size()]=-1;
			}
		}else{
			for (j=0;j<fiAry[i]->m_FoundCount;j++){
			
				if (fiAry[i]->m_isEqual[j]){
					snprintf(line,MAXSTRING,"%s%s%s %s %s%s%s",
						dir1,
						fiAry[i]->m_subdir,
						fiAry[i]->m_name,
						XRDS_ISACOPYOF,
						dir2,
						fiAry[i]->m_FoundSubdir[j],
						fiAry[i]->m_FoundName[j]);
				}else{
					snprintf(line,MAXSTRING,"@B3%s%s%s %s %s%s%s",
						dir1,
						fiAry[i]->m_subdir,
						fiAry[i]->m_name,
						XRDS_ISADIFFERENTCOPYOF,
						dir2,
						fiAry[i]->m_FoundSubdir[j],
						fiAry[i]->m_FoundName[j]);
				}					
				if ((!onlydiff)||(!fiAry[i]->m_isEqual[j])){
					(UI->filelist)->add(line);
					FileLine[(UI->filelist)->size()]=i;
					FoundLine[(UI->filelist)->size()]=j;			
				}
			}
		}
	}
}

// search files
void fillfilelist()
{
	fl_cursor(FL_CURSOR_WAIT);
	Fl::check();
	
	onlydiff=((UI->onlydiff)->value()==1);
	nocase=((UI->ignorecase)->value()==1);
	projmode=((UI->bProject)->value()==1);
	
	clearfilelist();
	parsefilter((UI->filter)->value());
	scandir((UI->finput)->value(),((UI->inputsubdir)->value()==1),"",true);
	scandir((UI->foutput)->value(),((UI->outputsubdir)->value()==1),"",false);
	printresult();

	fl_cursor(FL_CURSOR_DEFAULT);
	Fl::check();
}

// show complete diff output
// files=true in "files" mode
void showdiff(bool files)
{
	char file1[MAXSTRING]; // TODO dynamic
	char file2[MAXSTRING]; // TODO dynamic
	char label[MAXSTRING];
	
	if (!files){
		
		if (fiAryCount==0) return;
		
		int selected=(UI->filelist)->value();
		
		if (FoundLine[selected]==-1){
			fl_alert("%s%s %s",
				fiAry[FileLine[selected]]->m_subdir,
				fiAry[FileLine[selected]]->m_name,
				XRDS_NOTFOUND);
			return;
		}
			
		snprintf(file1,MAXSTRING,"%s/%s",
			fiAry[FileLine[selected]]->m_basedir,
			fiAry[FileLine[selected]]->m_name);
			
		snprintf(file2,MAXSTRING,"%s/%s",
			fiAry[FileLine[selected]]->m_FoundBasedir[FoundLine[selected]],
			fiAry[FileLine[selected]]->m_FoundName[FoundLine[selected]]);
			
		if (projmode){
			snprintf(label,MAXSTRING,"%s-1/%s%s -vs- %s-2/%s%s",
				XRDS_PROJECTDIR,
				fiAry[FileLine[selected]]->m_subdir,
				fiAry[FileLine[selected]]->m_name,
				XRDS_PROJECTDIR,
				fiAry[FileLine[selected]]->m_FoundSubdir[FoundLine[selected]],
				fiAry[FileLine[selected]]->m_FoundName[FoundLine[selected]]);
		}else{
			snprintf(label,MAXSTRING,"%s/%s%s -vs- %s/%s%s",
				XRDS_PROJECTDIR,
				fiAry[FileLine[selected]]->m_subdir,
				fiAry[FileLine[selected]]->m_name,
				XRDS_LIBDIR,
				fiAry[FileLine[selected]]->m_FoundSubdir[FoundLine[selected]],
				fiAry[FileLine[selected]]->m_FoundName[FoundLine[selected]]);
		}		
	}else{
		snprintf(file1,MAXSTRING,"%s",UI->finput->value());	
		snprintf(file2,MAXSTRING,"%s",UI->foutput->value());
		snprintf(label,MAXSTRING,"%s-1 -vs- %s-2",XRDS_FILE,XRDS_FILE);
	}	
	
	if (wDiff!=NULL){
		wDiff->hide();
		delete wDiff;
	}
	wDiff=new UIdiff(file1,file2,label);
}

int main(int argc, char* argv[]) 
{
	int next_option;
	char file1[MAXSTRING];
	char file2[MAXSTRING];
	*file1=0;
	*file2=0;

	const char* const short_options = "hV";
	const struct option long_options[] = {
	{ "help",     0, NULL, 'h' },
	{ "version",  0, NULL, 'V' },
	{ NULL,       0, NULL, 0   }
	};

	do {
		next_option = getopt_long (argc, argv, short_options,long_options, NULL);
		switch (next_option)
		{
			case 'h': 
				print_usage (0);
				break;
			case 'V':
				printf ("%s %s %s\n",PROGNAME,PROGVERS,AUTHORINFO);
				exit(0);
				break;
			case '?':
				print_usage (1);
				break;
		}
	}
	while (next_option != -1);
	
	if (optind < argc) snprintf(file1,MAXSTRING,argv[optind++]);
	if (optind < argc) snprintf(file2,MAXSTRING,argv[optind]);
	
	// start
	UI = new UImain();
	UI->make_window();
	
	if (*file1!=0) UI->finput->value(file1);
	if (*file2!=0) UI->foutput->value(file2);
	
	Fl::run();
	return 0;
}
